from django.http import HttpResponse


def encode_response(response):
    """
    Encodes a response to JSON-compatible datastructures
    """
    # TODO: Entirely useful things like cookies
    value = {
        "content_type": getattr(response, "content_type", None),
        "content": response.content,
        "status_code": response.status_code,
        "headers": response._headers.values(),
    }
    response.close()
    return value


def decode_response(value):
    """
    Decodes a response JSONish value to a HttpResponse object.
    """
    response = HttpResponse(
        content = value['content'],
        content_type = value['content_type'],
        status = value['status_code'],
    )
    response._headers = {k.lower: (k, v) for k, v in value['headers']}
    return response


class ResponseLater(Exception):
    """
    Class that represents a response which will be sent doown the response
    channel later. Used to move a django view-based segment onto the next
    task, as otherwise we'd need to write some kind of fake response.
    """
    pass



import time
import json
from collections import deque
from .base import BaseChannelBackend

queues = {}

class InMemoryChannelBackend(BaseChannelBackend):
    """
    In-memory channel implementation. Intended only for use with threading,
    in low-throughput development environments.
    """

    local_only = True

    def send(self, channel, message):
        # Try JSON encoding it to make sure it would, but store the native version
        json.dumps(message)
        # Add to the deque, making it if needs be
        queues.setdefault(channel, deque()).append(message)

    def receive_many(self, channels):
        if not channels:
            raise ValueError("Cannot receive on empty channel list!")
        # Try to pop a message from each channel
        for channel in channels:
            try:
                # This doesn't clean up empty channels - OK for testing.
                # For later versions, have cleanup w/lock.
                return channel, queues[channel].popleft()
            except (IndexError, KeyError):
                pass
        return None, None




import random
import string

from django.utils import six

from channels import channel_backends, DEFAULT_CHANNEL_BACKEND


class Channel(object):
    """
    Public interaction class for the channel layer.

    This is separate to the backends so we can:
     a) Hide receive_many from end-users, as it is only for interface servers
     b) Keep a stable-ish backend interface for third parties

    You can pass an alternate Channel Layer alias in, but it will use the
    "default" one by default.
    """

    def __init__(self, name, alias=DEFAULT_CHANNEL_BACKEND, channel_backend=None):
        """
        Create an instance for the channel named "name"
        """
        self.name = name
        if channel_backend:
            self.channel_backend = channel_backend
        else:
            self.channel_backend = channel_backends[alias]

    def send(self, **kwargs):
        """
        Send a message over the channel, taken from the kwargs.
        """
        self.channel_backend.send(self.name, kwargs)

    @classmethod
    def new_name(self, prefix):
        """
        Returns a new channel name that's unique and not closed
        with the given prefix. Does not need to be called before sending
        on a channel name - just provides a way to avoid clashing for
        response channels.
        """
        return "%s.%s" % (prefix, "".join(random.choice(string.ascii_letters) for i in range(32)))

    def as_view(self):
        """
        Returns a view version of this channel - one that takes
        the request passed in and dispatches it to our channel,
        serialized.
        """
        from channels.adapters import view_producer
        return view_producer(self.name)

    @classmethod
    def consumer(self, channels, alias=DEFAULT_CHANNEL_BACKEND):
        """
        Decorator that registers a function as a consumer.
        """
        # Upconvert if you just pass in a string
        if isinstance(channels, six.string_types):
            channels = [channels]
        # Get the channel 
        channel_backend = channel_backends[alias]
        # Return a function that'll register whatever it wraps
        def inner(func):
            channel_backend.registry.add_consumer(func, channels)
            return func
        return inner



import time
from django.core.management import BaseCommand, CommandError
from channels import channel_backends, DEFAULT_CHANNEL_BACKEND
from channels.interfaces.websocket_twisted import WebsocketTwistedInterface
from channels.utils import auto_import_consumers


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument('port', nargs='?',
            help='Optional port number')

    def handle(self, *args, **options):
        # Get the backend to use
        channel_backend = channel_backends[DEFAULT_CHANNEL_BACKEND]
        auto_import_consumers()
        if channel_backend.local_only:
            raise CommandError(
                "You have a process-local channel backend configured, and so cannot run separate interface servers.\n"
                "Configure a network-based backend in CHANNEL_BACKENDS to use this command."
            )
        # Run the interface
        port = options.get("port", None) or 9000
        self.stdout.write("Running Twisted/Autobahn WebSocket interface server")
        self.stdout.write(" Channel backend: %s" % channel_backend)
        self.stdout.write(" Listening on: ws://0.0.0.0:%i" % port)
        WebsocketTwistedInterface(channel_backend=channel_backend, port=port).run()



from django.http.request import HttpRequest
from django.http.response import HttpResponseBase
from django.core.handlers.base import BaseHandler
from .request import encode_request, decode_request
from .response import encode_response, decode_response, ResponseLater


def monkeypatch_django():
    """
    Monkeypatches support for us into parts of Django.
    """
    # Request encode/decode
    HttpRequest.channel_encode = encode_request
    HttpRequest.channel_decode = staticmethod(decode_request)
    # Response encode/decode
    HttpResponseBase.channel_encode = encode_response
    HttpResponseBase.channel_decode = staticmethod(decode_response)
    HttpResponseBase.ResponseLater = ResponseLater
    # Allow ResponseLater to propagate above handler
    BaseHandler.old_handle_uncaught_exception = BaseHandler.handle_uncaught_exception
    BaseHandler.handle_uncaught_exception = new_handle_uncaught_exception


def new_handle_uncaught_exception(self, request, resolver, exc_info):
    if exc_info[0] is ResponseLater:
        raise
    return BaseHandler.old_handle_uncaught_exception(self, request, resolver, exc_info)



import time
import json
import datetime
import redis
import uuid

from .base import BaseChannelBackend


class RedisChannelBackend(BaseChannelBackend):
    """
    ORM-backed channel environment. For development use only; it will span
    multiple processes fine, but it's going to be pretty bad at throughput.
    """

    def __init__(self, expiry=60, host="localhost", port=6379, prefix="django-channels:"):
        super(RedisChannelBackend, self).__init__(expiry)
        self.host = host
        self.port = port
        self.prefix = prefix

    @property
    def connection(self):
        """
        Returns the correct connection for the current thread.
        """
        return redis.Redis(host=self.host, port=self.port)

    def send(self, channel, message):
        # Write out message into expiring key (avoids big items in list)
        key = uuid.uuid4()
        self.connection.set(
            key,
            json.dumps(message),
        )
        self.connection.expire(
            key,
            self.expiry + 10,
        )
        # Add key to list
        self.connection.rpush(
            self.prefix + channel,
            key,
        )
        # Set list to expire when message does (any later messages will bump this)
        self.connection.expire(
            self.prefix + channel,
            self.expiry + 10,
        )
        # TODO: Prune expired messages from same list (in case nobody consumes)

    def receive_many(self, channels):
        if not channels:
            raise ValueError("Cannot receive on empty channel list!")
        # Get a message from one of our channels
        while True:
            result = self.connection.blpop([self.prefix + channel for channel in channels], timeout=1)
            if result:
                content = self.connection.get(result[1])
                if content is None:
                    continue
                return result[0][len(self.prefix):], json.loads(content)
            else:
                return None, None

    def __str__(self):
        return "%s(host=%s, port=%s)" % (self.__class__.__name__, self.host, self.port)



import time
from wsgiref.simple_server import BaseHTTPRequestHandler
from django.core.management import BaseCommand, CommandError
from channels import channel_backends, DEFAULT_CHANNEL_BACKEND
from channels.worker import Worker
from channels.utils import auto_import_consumers


class Command(BaseCommand):

    def handle(self, *args, **options):
        # Get the backend to use
        channel_backend = channel_backends[DEFAULT_CHANNEL_BACKEND]
        auto_import_consumers()
        if channel_backend.local_only:
            raise CommandError(
                "You have a process-local channel backend configured, and so cannot run separate workers.\n"
                "Configure a network-based backend in CHANNEL_BACKENDS to use this command."
            )
        # Launch a worker
        self.stdout.write("Running worker against backend %s" % channel_backend)
        # Optionally provide an output callback
        callback = None
        if options.get("verbosity", 1) > 1:
            callback = self.consumer_called
        # Run the worker
        try:
            Worker(channel_backend=channel_backend, callback=callback).run()
        except KeyboardInterrupt:
            pass

    def consumer_called(self, channel, message):
        self.stdout.write("[%s] %s" % (self.log_date_time_string(), channel))

    def log_date_time_string(self):
        """Return the current time formatted for logging."""
        now = time.time()
        year, month, day, hh, mm, ss, x, y, z = time.localtime(now)
        s = "%02d/%3s/%04d %02d:%02d:%02d" % (
                day, BaseHTTPRequestHandler.monthname[month], year, hh, mm, ss)
        return s



from django.http import HttpRequest
from django.utils.datastructures import MultiValueDict


def encode_request(request):
    """
    Encodes a request to JSON-compatible datastructures
    """
    # TODO: More stuff
    value = {
        "GET": request.GET.items(),
        "POST": request.POST.items(),
        "COOKIES": request.COOKIES,
        "META": {k: v for k, v in request.META.items() if not k.startswith("wsgi")},
        "path": request.path,
        "path_info": request.path_info,
        "method": request.method,
        "response_channel": request.response_channel,
    }
    return value


def decode_request(value):
    """
    Decodes a request JSONish value to a HttpRequest object.
    """
    request = HttpRequest()
    request.GET = MultiValueDict(value['GET'])
    request.POST = MultiValueDict(value['POST'])
    request.COOKIES = value['COOKIES']
    request.META = value['META']
    request.path = value['path']
    request.method = value['method']
    request.path_info = value['path_info']
    request.response_channel = value['response_channel']
    return request



import functools

from django.utils import six

from .utils import name_that_thing


class ConsumerRegistry(object):
    """
    Manages the available consumers in the project and which channels they
    listen to.

    Generally this is attached to a backend instance as ".registry"
    """

    def __init__(self):
        self.consumers = {}

    def add_consumer(self, consumer, channels):
        # Upconvert if you just pass in a string
        if isinstance(channels, six.string_types):
            channels = [channels]
        # Register on each channel, checking it's unique
        for channel in channels:
            if channel in self.consumers:
                raise ValueError("Cannot register consumer %s - channel %r already consumed by %s" % (
                    name_that_thing(consumer),
                    channel,
                    name_that_thing(self.consumers[channel]),
                ))
            self.consumers[channel] = consumer

    def all_channel_names(self):
        return self.consumers.keys()

    def consumer_for_channel(self, channel):
        try:
            return self.consumers[channel]
        except KeyError:
            return None



import time


class Worker(object):
    """
    A "worker" process that continually looks for available messages to run
    and runs their consumers.
    """

    def __init__(self, channel_backend, callback=None):
        self.channel_backend = channel_backend
        self.callback = callback

    def run(self):
        """
        Tries to continually dispatch messages to consumers.
        """
        channels = self.channel_backend.registry.all_channel_names()
        while True:
            channel, message = self.channel_backend.receive_many_blocking(channels)
            # Handle the message
            consumer = self.channel_backend.registry.consumer_for_channel(channel)
            if self.callback:
                self.callback(channel, message)
            consumer(channel=channel, **message)



import django
import time
from collections import deque
from twisted.internet import reactor
from channels import Channel, channel_backends, DEFAULT_CHANNEL_BACKEND
from autobahn.twisted.websocket import WebSocketServerProtocol, WebSocketServerFactory


class InterfaceProtocol(WebSocketServerProtocol):
    """
    Protocol which supports WebSockets and forwards incoming messages to
    the django.websocket channels.
    """

    def onConnect(self, request):
        self.channel_backend = channel_backends[DEFAULT_CHANNEL_BACKEND]
        self.request_info = {
            "path": request.path,
        }

    def onOpen(self):
        # Make sending channel
        self.send_channel = Channel.new_name("django.websocket.send")
        self.factory.protocols[self.send_channel] = self
        # Send news that this channel is open
        Channel("django.websocket.connect").send(
            send_channel = self.send_channel,
            **self.request_info
        )

    def onMessage(self, payload, isBinary):
        if isBinary:
            Channel("django.websocket.receive").send(
                send_channel = self.send_channel,
                content = payload,
                binary = True,
                **self.request_info
            )
        else:
            Channel("django.websocket.receive").send(
                send_channel = self.send_channel,
                content = payload.decode("utf8"),
                binary = False,
                **self.request_info
            )

    def serverSend(self, content, binary=False, **kwargs):
        """
        Server-side channel message to send a message.
        """
        if binary:
            self.sendMessage(content, binary)
        else:
            self.sendMessage(content.encode("utf8"), binary)

    def serverClose(self):
        """
        Server-side channel message to close the socket
        """
        self.sendClose()

    def onClose(self, wasClean, code, reason):
        del self.factory.protocols[self.send_channel]
        Channel("django.websocket.disconnect").send(
            send_channel = self.send_channel,
            **self.request_info
        )


class InterfaceFactory(WebSocketServerFactory):
    """
    Factory which keeps track of its open protocols' receive channels
    and can dispatch to them.
    """

    # TODO: Clean up dead protocols if needed?

    def __init__(self, *args, **kwargs):
        super(InterfaceFactory, self).__init__(*args, **kwargs)
        self.protocols = {}

    def send_channels(self):
        return self.protocols.keys()

    def dispatch_send(self, channel, message):
        if message.get("close", False):
            self.protocols[channel].serverClose()
        else:
            self.protocols[channel].serverSend(**message)


class WebsocketTwistedInterface(object):
    """
    Easy API to run a WebSocket interface server using Twisted.
    Integrates the channel backend by running it in a separate thread, using
    the always-compatible polling style.
    """

    def __init__(self, channel_backend, port=9000):
        self.channel_backend = channel_backend
        self.port = port

    def run(self):
        self.factory = InterfaceFactory("ws://0.0.0.0:%i" % self.port, debug=False)
        self.factory.protocol = InterfaceProtocol
        reactor.listenTCP(self.port, self.factory)
        reactor.callInThread(self.backend_reader)
        reactor.run()

    def backend_reader(self):
        """
        Run in a separate thread; reads messages from the backend.
        """
        while True:
            channels = self.factory.send_channels()
            # Quit if reactor is stopping
            if not reactor.running:
                return
            # Don't do anything if there's no channels to listen on
            if channels:
                channel, message = self.channel_backend.receive_many(channels)
            else:
                time.sleep(0.1)
                continue
            # Wait around if there's nothing received
            if channel is None:
                time.sleep(0.05)
                continue
            # Deal with the message
            self.factory.dispatch_send(channel, message)



__version__ = "0.1.1"

# Load backends, using settings if available (else falling back to a default)
DEFAULT_CHANNEL_BACKEND = "default"
from .backends import BackendManager
from django.conf import settings
channel_backends = BackendManager(
    getattr(settings, "CHANNEL_BACKENDS", {
        DEFAULT_CHANNEL_BACKEND: {
            "BACKEND": "channels.backends.memory.InMemoryChannelBackend",
        }
    })
)

# Ensure monkeypatching
from .hacks import monkeypatch_django
monkeypatch_django()

# Promote channel to top-level (down here to avoid circular import errs)
from .channel import Channel



import functools

from django.core.handlers.base import BaseHandler
from django.http import HttpRequest, HttpResponse

from channels import Channel, channel_backends, DEFAULT_CHANNEL_BACKEND


class UrlConsumer(object):
    """
    Dispatches channel HTTP requests into django's URL system.
    """

    def __init__(self):
        self.handler = BaseHandler()
        self.handler.load_middleware()

    def __call__(self, channel, **kwargs):
        request = HttpRequest.channel_decode(kwargs)
        try:
            response = self.handler.get_response(request)
        except HttpResponse.ResponseLater:
            return
        Channel(request.response_channel).send(**response.channel_encode())


def view_producer(channel_name):
    """
    Returns a new view function that actually writes the request to a channel
    and abandons the response (with an exception the Worker will catch)
    """
    def producing_view(request):
        Channel(channel_name).send(**request.channel_encode())
        raise HttpResponse.ResponseLater()
    return producing_view


def view_consumer(channel_name, alias=DEFAULT_CHANNEL_BACKEND):
    """
    Decorates a normal Django view to be a channel consumer.
    Does not run any middleware
    """
    def inner(func): 
        @functools.wraps(func)
        def consumer(channel, **kwargs):
            request = HttpRequest.channel_decode(kwargs)
            response = func(request)
            Channel(request.response_channel).send(**response.channel_encode())
        # Get the channel layer and register
        channel_backend = channel_backends[DEFAULT_CHANNEL_BACKEND]
        channel_backend.registry.add_consumer(consumer, [channel_name])
        return func
    return inner



import django
import threading
from django.core.management.commands.runserver import Command as RunserverCommand
from django.core.management import CommandError
from channels import channel_backends, DEFAULT_CHANNEL_BACKEND
from channels.worker import Worker
from channels.utils import auto_import_consumers
from channels.adapters import UrlConsumer
from channels.interfaces.wsgi import WSGIInterface


class Command(RunserverCommand):

    def get_handler(self, *args, **options):
        """
        Returns the default WSGI handler for the runner.
        """
        return WSGIInterface(self.channel_backend)

    def run(self, *args, **options):
        # Force disable reloader for now
        options['use_reloader'] = False
        # Check a handler is registered for http reqs
        self.channel_backend = channel_backends[DEFAULT_CHANNEL_BACKEND]
        auto_import_consumers()
        if not self.channel_backend.registry.consumer_for_channel("django.wsgi.request"):
            # Register the default one
            self.channel_backend.registry.add_consumer(UrlConsumer(), ["django.wsgi.request"])
        # Launch a worker thread
        worker = WorkerThread(self.channel_backend)
        worker.daemon = True
        worker.start()
        # Run the rest
        return super(Command, self).run(*args, **options)


class WorkerThread(threading.Thread):
    """
    Class that runs a worker
    """

    def __init__(self, channel_backend):
        super(WorkerThread, self).__init__()
        self.channel_backend = channel_backend

    def run(self):
        Worker(channel_backend=self.channel_backend).run()



import time
import json
import datetime

from django.apps.registry import Apps
from django.db import models, connections, DEFAULT_DB_ALIAS
from django.utils.functional import cached_property
from django.utils.timezone import now

from .base import BaseChannelBackend


class DatabaseChannelBackend(BaseChannelBackend):
    """
    ORM-backed channel environment. For development use only; it will span
    multiple processes fine, but it's going to be pretty bad at throughput.
    """

    def __init__(self, expiry=60, db_alias=DEFAULT_DB_ALIAS):
        super(DatabaseChannelBackend, self).__init__(expiry)
        self.db_alias = db_alias

    @property
    def connection(self):
        """
        Returns the correct connection for the current thread.
        """
        return connections[self.db_alias]

    @property
    def model(self):
        """
        Initialises a new model to store messages; not done as part of a
        models.py as we don't want to make it for most installs.
        """
        # Make the model class
        class Message(models.Model):
            # We assume an autoincrementing PK for message order
            channel = models.CharField(max_length=200, db_index=True)
            content = models.TextField()
            expiry = models.DateTimeField(db_index=True)
            class Meta:
                apps = Apps()
                app_label = "channels"
                db_table = "django_channels"
        # Ensure its table exists
        if Message._meta.db_table not in self.connection.introspection.table_names(self.connection.cursor()):
            with self.connection.schema_editor() as editor:
                editor.create_model(Message)
        return Message

    def send(self, channel, message):
        self.model.objects.create(
            channel = channel,
            content = json.dumps(message),
            expiry = now() + datetime.timedelta(seconds=self.expiry)
        )

    def receive_many(self, channels):
        if not channels:
            raise ValueError("Cannot receive on empty channel list!")
        # Delete all expired messages (add 10 second grace period for clock sync)
        self.model.objects.filter(expiry__lt=now() - datetime.timedelta(seconds=10)).delete()
        # Get a message from one of our channels
        message = self.model.objects.filter(channel__in=channels).order_by("id").first()
        if message:
            self.model.objects.filter(pk=message.pk).delete()
            return message.channel, json.loads(message.content)
        else:
            return None, None

    def __str__(self):
        return "%s(alias=%s)" % (self.__class__.__name__, self.connection.alias)



import time
from channels.consumer_registry import ConsumerRegistry


class ChannelClosed(Exception):
    """
    Raised when you try to send to a closed channel.
    """
    pass


class BaseChannelBackend(object):
    """
    Base class for all channel layer implementations. Manages both sending
    and receving messages from the backend, and each comes with its own
    registry of consumers.
    """

    # Flags if this backend can only be used inside one process.
    # Causes errors if you try to run workers/interfaces separately with it.
    local_only = False

    def __init__(self, expiry=60):
        self.registry = ConsumerRegistry()
        self.expiry = expiry

    def send(self, channel, message):
        """
        Send a message over the channel, taken from the kwargs.
        """
        raise NotImplementedError()

    def receive_many(self, channels):
        """
        Return the first message available on one of the
        channels passed, as a (channel, message) tuple, or return (None, None)
        if no channels are available.

        Should not block, but is allowed to be moderately slow/have a short
        timeout - it needs to return so we can refresh the list of channels,
        not because the rest of the process is waiting on it.

        Better performance can be achieved for interface servers by directly
        integrating the server and the backend code; this is merely for a
        generic support-everything pattern.
        """
        raise NotImplementedError()

    def receive_many_blocking(self, channels):
        """
        Blocking version of receive_many, if the calling context knows it
        doesn't ever want to change the channels list until something happens.

        This base class provides a default implementation; can be overridden
        to be more efficient by subclasses.
        """
        while True:
            channel, message = self.receive_many(channels)
            if channel is None:
                time.sleep(0.05)
                continue
            return channel, message

    def __str__(self):
        return self.__class__.__name__



import types
from django.apps import apps


def auto_import_consumers():
    """
    Auto-import consumers modules in apps
    """
    for app_config in apps.get_app_configs():
        for submodule in ["consumers", "views"]:
            module_name = "%s.%s" % (app_config.name, submodule)
            try:
                __import__(module_name)
            except ImportError as e:
                if "no module named %s" % submodule not in str(e).lower():
                    raise


def name_that_thing(thing):
    """
    Returns either the function/class path or just the object's repr
    """
    if hasattr(thing, "__name__"):
        if hasattr(thing, "__class__") and not isinstance(thing, types.FunctionType):
            if thing.__class__ is not type:
                return name_that_thing(thing.__class__)
        if hasattr(thing, "__module__"):
            return "%s.%s" % (thing.__module__, thing.__name__)
    return repr(thing)



from django.utils.module_loading import import_string


class InvalidChannelBackendError(ValueError):
    pass


class BackendManager(object):
    """
    Takes a settings dictionary of backends and initialises them.
    """

    def __init__(self, backend_configs):
        self.configs = backend_configs
        self.backends = {}
            
    def make_backend(self, name):
        # Load the backend class
        try:
            backend_class = import_string(self.configs[name]['BACKEND'])
        except KeyError:
            raise InvalidChannelBackendError("No BACKEND specified for %s" % name)
        except ImportError as e:
            raise InvalidChannelBackendError("Cannot import BACKEND %r specified for %s" % (self.configs[name]['BACKEND'], name))
        # Initialise and pass config
        instance = backend_class(**{k.lower(): v for k, v in self.configs[name].items() if k != "BACKEND"})
        instance.alias = name
        return instance

    def __getitem__(self, key):
        if key not in self.backends:
            self.backends[key] = self.make_backend(key)
        return self.backends[key]



import django
from django.core.handlers.wsgi import WSGIHandler
from django.http import HttpResponse
from channels import Channel


class WSGIInterface(WSGIHandler):
    """
    WSGI application that pushes requests to channels.
    """

    def __init__(self, channel_backend, *args, **kwargs):
        self.channel_backend = channel_backend
        django.setup()
        super(WSGIInterface, self).__init__(*args, **kwargs)

    def get_response(self, request):
        request.response_channel = Channel.new_name("django.wsgi.response")
        Channel("django.wsgi.request", channel_backend=self.channel_backend).send(**request.channel_encode())
        channel, message = self.channel_backend.receive_many_blocking([request.response_channel])
        return HttpResponse.channel_decode(message)
